package myActions;

import com.intellij.debugger.DebuggerManagerEx;
import com.intellij.debugger.engine.DebuggerManagerThreadImpl;
import com.intellij.debugger.impl.DebuggerContextImpl;
import com.intellij.debugger.jdi.DecompiledLocalVariable;
import com.intellij.debugger.jdi.StackFrameProxyImpl;
import com.intellij.openapi.actionSystem.AnAction;
import com.intellij.openapi.actionSystem.AnActionEvent;
import com.intellij.openapi.actionSystem.CommonDataKeys;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.Messages;
import com.sun.jdi.InternalException;
import com.sun.jdi.StackFrame;
import com.sun.jdi.Value;
import com.sun.jdi.VirtualMachine;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.lang.reflect.*;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.*;

public class GetAdviceAction extends AnAction
{
    private static final boolean ourInitializationOk;
    private static Class<?> ourSlotInfoClass;
    private static Constructor<?> slotInfoConstructor;
    private static Class<?> ourGetValuesClass;
    private static Method ourEnqueueMethod;
    private static Method ourWaitForReplyMethod;

    static {
        boolean success = false;
        try {
            ourSlotInfoClass = Class.forName("com.sun.tools.jdi.JDWP$StackFrame$GetValues$SlotInfo");
            slotInfoConstructor = ourSlotInfoClass.getDeclaredConstructor(int.class, byte.class);
            slotInfoConstructor.setAccessible(true);

            ourGetValuesClass = Class.forName("com.sun.tools.jdi.JDWP$StackFrame$GetValues");
            ourEnqueueMethod = findMethod(ourGetValuesClass, "enqueueCommand");
            ourEnqueueMethod.setAccessible(true);
            ourWaitForReplyMethod = findMethod(ourGetValuesClass, "waitForReply");
            ourWaitForReplyMethod.setAccessible(true);

            success = true;
        }
        catch (Throwable e) {
            e.printStackTrace();
        }
        ourInitializationOk = success;
    }

    private static Method findMethod(Class aClass, String methodName) throws NoSuchMethodException {
        for (Method method : aClass.getDeclaredMethods()) {
            if (methodName.equals(method.getName())) {
                return method;
            }
        }
        throw new NoSuchMethodException(aClass.getName() + "." + methodName);
    }

    public static Map<DecompiledLocalVariable, Value> fetchValues(StackFrame frame, Collection<DecompiledLocalVariable> vars, Project project) throws Exception {
        if (!ourInitializationOk) {
            return Collections.emptyMap();
        }
        final Field frameIdField = frame.getClass().getDeclaredField("id");
        frameIdField.setAccessible(true);
        final Object frameId = frameIdField.get(frame);

        final VirtualMachine vm = frame.virtualMachine();
        final Method stateMethod = vm.getClass().getDeclaredMethod("state");
        stateMethod.setAccessible(true);

        Object slotInfoArray = createSlotInfoArray(vars);

        Object ps;
        final Object vmState = stateMethod.invoke(vm);
        synchronized(vmState) {
            ps = ourEnqueueMethod.invoke(null, vm, frame.thread(), frameId, slotInfoArray);
        }

        try {
            System.out.println(((DebuggerManagerThreadImpl.isManagerThread()) ? "In " : "Not in ") + " Manager thread");
            // Getting the following message in the Event Log of the debugee (i.e. IntelliJ CE) at runtime AND in the local debugger console:
            // Assertion failed: Should be invoked in manager thread, use DebuggerManagerThreadImpl.getInstance(..).invoke...
            final Object reply = ourWaitForReplyMethod.invoke(null, vm, ps);
            final Field valuesField = reply.getClass().getDeclaredField("values");
            valuesField.setAccessible(true);
            final Value[] values = (Value[])valuesField.get(reply);
            if (vars.size() != values.length) {
                throw new InternalException("Wrong number of values returned from target VM");
            }
            final Map<DecompiledLocalVariable, Value> map = new HashMap<DecompiledLocalVariable, Value>(vars.size());
            int idx = 0;
            for (DecompiledLocalVariable var : vars) {
                map.put(var, values[idx++]);
            }
            return map;
        }
        catch (InvocationTargetException e) {
            final Throwable target = e.getTargetException();
            if (target instanceof Exception) {
                throw (Exception)target;
            }
            throw e;
        }
    }

    private static Object createSlotInfoArray(Collection<DecompiledLocalVariable> vars) throws Exception {
        final Object arrayInstance = Array.newInstance(ourSlotInfoClass, vars.size());

        int idx = 0;
        for (DecompiledLocalVariable var : vars) {
            final Object info = slotInfoConstructor.newInstance(var.getSlot(), (byte)var.getSignature().charAt(0));
            Array.set(arrayInstance, idx++, info);
        }

        return arrayInstance;
    }

      public void actionPerformed(AnActionEvent event)
      {

        StringBuilder sb = new StringBuilder();
          try {

              Project project = event.getData(CommonDataKeys.PROJECT);
              DebuggerContextImpl dbgContext = (DebuggerManagerEx.getInstanceEx(project)).getContext();
              StackFrameProxyImpl sfProxy = dbgContext.getFrameProxy();
              StackFrame frame = sfProxy.getStackFrame();

              Collection<DecompiledLocalVariable> vars = new ArrayList<DecompiledLocalVariable>();
              vars.add(new DecompiledLocalVariable(0, "headers", "String[]"));
              vars.add(new DecompiledLocalVariable(1, "args", "String[]"));

              Map localVars = fetchValues(frame, vars, project); // LocalVariablesUtil.fetchValues(frame, vars);

              URL url = new URL("http://localhost:8080/HeadsUp/main/index");
              HttpURLConnection conn = (HttpURLConnection) url.openConnection();
              conn.setDoOutput(true);
              conn.setRequestMethod("POST");
              conn.setRequestProperty("Content-Type", "application/json");

              String input = "{\"headers\":[\"name\":\"Accept\",\"value\":\"text/html\"]}";

              OutputStream os = conn.getOutputStream();
              os.write(input.getBytes());
              os.flush();
              BufferedReader br = new BufferedReader(new InputStreamReader((conn.getInputStream())));

              String output;
              System.out.println("Output from Server .... \n");
              while ((output = br.readLine()) != null) {
                  System.out.println(output);
                  sb.append(output);
              }

              conn.disconnect();

          } catch (Exception e) {

              e.printStackTrace();

          }

        Messages.showMessageDialog(sb.toString(), "Information", Messages.getInformationIcon());
      }
}
